@@ -26,7 +26,16 @@ namespace Decibel {
 /** "Special" names for use in QVariantMaps: */
 const QString name_protocol("decibel_protocol");
 const QString name_presence("decibel_presence");
+const QString name_display_name("decibel_display_name");
+const QString name_current_presence("decibel_current_presence");
 const QString name_presence_message("decibel_presence_message");
+const QString name_autoreconnect("decibel_auto_reconnect");
+
+/** "Normal" Telepathy names: */
+const QString name_tp_server("server");
+const QString name_tp_account("account");
+const QString name_tp_passwort("passwort");
+const QString name_tp_port("port");
 
 } // namespace
 
