@@ -39,6 +39,8 @@ namespace
     static const QString account_array("Accounts");
 }
 
+/// @cond INCLUDE_PRIVATE
+
 /**
  * @brief Private class implementing the AccountManager.
  *
@@ -54,7 +56,8 @@ public:
      */
     AccountManagerPrivate(ConnectionFacade * conn_mgr) :
         m_currentHandle(1),
-        m_connectionFacade(conn_mgr)
+        m_connectionFacade(conn_mgr),
+        blockReconnects(false)
     {
         Q_ASSERT(0 != conn_mgr);
 
@@ -77,11 +80,8 @@ public:
                 { value = QVariant(value.toInt()); }
                 account_data.insert(current_key, value);
             }
-            int check_result = checkNVPairs(account_data);
-            if (check_result < 0)
-            { qWarning() << "Account data invalid:" << check_result; }
-            else
-            { createAccount(account_data); }
+            account_data = sanityCheck(account_data);
+            createAccount(account_data);
         }
         settings.endArray();
         settings.endGroup();
@@ -126,41 +126,17 @@ public:
      */
     uint createAccount(const QVariantMap & nv_pairs)
     {
-        Q_ASSERT(nv_pairs.size() > 2);
+        Q_ASSERT(nv_pairs.contains(Decibel::name_display_name));
         Q_ASSERT(nv_pairs.contains(Decibel::name_protocol));
         Q_ASSERT(nv_pairs.contains(Decibel::name_presence));
         Q_ASSERT(nv_pairs.contains(Decibel::name_presence_message));
+        Q_ASSERT(nv_pairs.contains(Decibel::name_current_presence));
+        Q_ASSERT(nv_pairs.contains(Decibel::name_autoreconnect));
 
         m_accounts[m_currentHandle] = nv_pairs;
         return m_currentHandle++;
     }
 
-    /**
-     * @brief Check the data for completeness.
-     * @param nv_pairs The account data.
-     * @return 0 on success or a negative error code.
-     *
-     * Check the account data for completeness and return an error code if
-     * it is not.
-     */
-    int checkNVPairs(const QVariantMap & nv_pairs) const
-    {
-        // Fail on empty data:
-        if (nv_pairs.isEmpty())
-        { return Decibel::ERROR_EMPTY_LIST; }
-
-        if (!nv_pairs.contains(Decibel::name_protocol))
-        { return Decibel::ERROR_NO_PROTOCOL; }
-
-        if (QString(nv_pairs[Decibel::name_protocol].typeName()) != QString("QString"))
-        { return Decibel::ERROR_WRONG_VALUE_TYPE_FOR_PROTOCOL; }
-
-        if (nv_pairs.uniqueKeys().size() != nv_pairs.keys().size())
-        { return Decibel::ERROR_DUPLICATE_NAME; }
-
-        return 0;
-    }
-
     /**
      * @brief Make sure a account has the required presence information.
      * @param nv_pairs The account data.
@@ -169,10 +145,13 @@ public:
      * Make sure a account has the required presence information and fill in
      * the blanks if there are any.
      */
-    QVariantMap ensurePresenceIn(const QVariantMap & nv_pairs) const
+    QVariantMap sanityCheck(const QVariantMap & nv_pairs) const
     {
         QVariantMap pairs(nv_pairs);
 
+        if (!pairs.contains(Decibel::name_protocol))
+        { return QVariantMap(); }
+
         if (!pairs.contains(Decibel::name_presence))
         { pairs.insert(Decibel::name_presence, QVariant(QtTapioca::ContactBase::Offline)); }
         else
@@ -184,11 +163,45 @@ public:
         }
         if (!pairs.contains(Decibel::name_presence_message))
         { pairs.insert(Decibel::name_presence_message, QVariant(QString())); }
+        if (!pairs.contains(Decibel::name_autoreconnect))
+        { pairs.insert(Decibel::name_autoreconnect, QVariant(false)); }
+        if (!pairs.contains(Decibel::name_display_name))
+        {
+            // create a sensible display name:
+            QString protocol;
+            QString account;
+            QString server;
+            QString port;
 
+            if (pairs.contains(Decibel::name_protocol))
+            { protocol = pairs[Decibel::name_protocol].toString(); }
+            if (pairs.contains(Decibel::name_tp_account))
+            { account = pairs[Decibel::name_tp_account].toString(); }
+            if (pairs.contains(Decibel::name_tp_server))
+            { server = pairs[Decibel::name_tp_server].toString(); }
+            if (pairs.contains(Decibel::name_tp_port))
+            { port = pairs[Decibel::name_tp_port].toString(); }
+
+            if (protocol.isEmpty()) { protocol = "unknown"; }
+            if (account.isEmpty()) { account = "unknown"; }
+            if (server.isEmpty()) { server = "unknown"; }
+            if (!port.isEmpty()) { port = ':' + port; }
+
+            QString name(protocol + "://" +
+                         account + '@' + server + port + '/');
+            pairs.insert(Decibel::name_display_name, QVariant(name));
+        }
+
+        pairs.insert(Decibel::name_current_presence, QVariant(QtTapioca::ContactBase::Offline));
+
+        Q_ASSERT(pairs.contains(Decibel::name_protocol));
         Q_ASSERT(pairs.contains(Decibel::name_presence));
+        Q_ASSERT(pairs.contains(Decibel::name_current_presence));
+        Q_ASSERT(pairs.contains(Decibel::name_autoreconnect));
         Q_ASSERT(pairs.contains(Decibel::name_presence_message));
+        Q_ASSERT(pairs.contains(Decibel::name_display_name));
         Q_ASSERT(pairs.size() >= nv_pairs.size() &&
-                 pairs.size() <= nv_pairs.size() + 2 );
+                 pairs.size() <= nv_pairs.size() + 5 );
         return pairs;
     }
 
@@ -216,7 +229,8 @@ public:
         if (presence_state == QtTapioca::ContactBase::Offline)
         {
             // We need to disconnect:
-            if (0 != connection) { connection->disconnect(); }
+            if (0 != connection)
+            { connection->disconnect(); }
             // connection pointer is set on signal!
         }
         else
@@ -227,7 +241,9 @@ public:
                 connection =
                     m_connectionFacade->connectUsingAccount(m_accounts[account_handle].data, presence_state, message);
                 if (connection == 0)
-                { result = -1; /* FIXME: Give proper error code! */ }
+                {
+                    return QtTapioca::ContactBase::Offline; /* FIXME: Give proper error code! */
+                }
                 setConnection(account_handle, connection);
             }
             else
@@ -251,6 +267,8 @@ public:
     void setConnection(const uint account_handle,
                        QtTapioca::Connection * connection)
     {
+        Q_ASSERT(0 == connectionOf(account_handle) ||
+                 0 == connection);
         if (!gotAccount(account_handle)) { return; }
         m_accounts[account_handle].connection = connection;
     }
@@ -305,15 +323,38 @@ public:
      * Update presence information on an account, bringing it on/off line as
      * required.
      */
-    void updatePresence(const uint account_handle,
+    void updatePresenceLocally(const uint account_handle,
                         const QtTapioca::ContactBase::Presence presence_state,
                         const QString message)
     {
         Q_ASSERT(gotAccount(account_handle));
+        Q_ASSERT(m_accounts[account_handle].data.contains(Decibel::name_current_presence));
         Q_ASSERT(m_accounts[account_handle].data.contains(Decibel::name_presence));
-        m_accounts[account_handle].data[Decibel::name_presence] = QVariant(presence_state);
+        Q_ASSERT(m_accounts[account_handle].data.contains(Decibel::name_autoreconnect));
+        Q_ASSERT(m_accounts[account_handle].data.contains(Decibel::name_presence_message));
+
+        m_accounts[account_handle].data[Decibel::name_current_presence] =
+            QVariant(presence_state);
         if (!message.isEmpty())
-        { m_accounts[account_handle].data[Decibel::name_presence_message] = QVariant(message); }
+        {
+            m_accounts[account_handle].data[Decibel::name_presence_message] =
+                QVariant(message);
+        }
+
+        // Get back to intended presence state:
+        if (blockReconnects) { return; }
+
+        if (m_accounts[account_handle].data[Decibel::name_autoreconnect].toBool())
+        {
+            const QtTapioca::ContactBase::Presence intended_state =
+                static_cast<QtTapioca::ContactBase::Presence>(
+                m_accounts[account_handle].data[Decibel::name_presence].toInt());
+            if (intended_state != presence_state)
+            {
+                manageConnection(account_handle, intended_state,
+                    m_accounts[account_handle].data[Decibel::name_presence_message].toString());
+            }
+        }
     }
 
     /**
@@ -323,7 +364,7 @@ public:
      *
      * Update the presence message on an account.
      */
-    void updatePresenceMessage(const uint account_handle, const QString & message)
+    void updatePresenceMessageLocally(const uint account_handle, const QString & message)
     {
         Q_ASSERT(gotAccount(account_handle));
         m_accounts[account_handle].data[Decibel::name_presence_message] = QVariant(message);
@@ -374,23 +415,27 @@ public:
     /** @brief A pointer to the D-Bus Adaptor of the AccountManager. */
     QPointer<AccountManagerAdaptor> m_adaptor;
 
+    bool blockReconnects;
+
     friend class AccountManager;
 };
 
+/// @endcond
+
 // ****************************************************************************
 
-AccountManager::AccountManager(ConnectionFacade * connection_mgr, QObject * parent) :
+AccountManager::AccountManager(ConnectionFacade * connection_facade, QObject * parent) :
     QObject(parent),
-    d(new AccountManagerPrivate(connection_mgr))
+    d(new AccountManagerPrivate(connection_facade))
 {
     Q_ASSERT(d != 0);
     d->m_adaptor = new AccountManagerAdaptor(this);
     Q_ASSERT(d->m_adaptor != 0);
 
-    connect(connection_mgr,
+    connect(connection_facade,
         SIGNAL(ownPresenceUpdated(QtTapioca::Connection *, const int, const QString)),
         SLOT(onOwnPresenceUpdated(QtTapioca::Connection *, const int, const QString)));
-    connect(connection_mgr,
+    connect(connection_facade,
         SIGNAL(connectionClosed(QtTapioca::Connection *)),
         SLOT(onConnectionClosed(QtTapioca::Connection *)));
 }
@@ -403,22 +448,26 @@ AccountManager::~AccountManager()
 void AccountManager::bringUpAccounts()
 {
     uint i;
-    qDebug() << "Bringing up accounts...";
     foreach(i, d->m_accounts.keys())
     {
         Q_ASSERT(d->m_accounts[i].data.contains(Decibel::name_protocol));
+        Q_ASSERT(d->m_accounts[i].data.contains(Decibel::name_current_presence));
         Q_ASSERT(d->m_accounts[i].data.contains(Decibel::name_presence));
         Q_ASSERT(d->m_accounts[i].data.contains(Decibel::name_presence_message));
+        Q_ASSERT(d->m_accounts[i].data.contains(Decibel::name_autoreconnect));
+
+        qDebug() << "Bringing up" << i << ":"
+                 << d->m_accounts[i].data[Decibel::name_protocol]
+                 << "ip:" << d->m_accounts[i].data[Decibel::name_presence]
+                 << "cp:"
+                 << d->m_accounts[i].data[Decibel::name_current_presence]
+                 << "pm:"
+                 << d->m_accounts[i].data[Decibel::name_presence_message];
 
         setPresenceAndMessage(i,
                               d->m_accounts[i].data[Decibel::name_presence].toInt(),
                               d->m_accounts[i].data[Decibel::name_presence_message].toString());
-        qDebug() << "    " << i << ":"
-                 << d->m_accounts[i].data[Decibel::name_protocol]
-                 << d->m_accounts[i].data[Decibel::name_presence]
-                 << d->m_accounts[i].data[Decibel::name_presence_message];
     }
-    qDebug() << "All accounts updated.";
 }
 
 QString AccountManager::protocol(const uint account_handle) const
@@ -453,13 +502,15 @@ QVariantMap AccountManager::queryAccount(const uint account_handle) const
 uint AccountManager::addAccount(const QVariantMap & nv_pairs)
 {
     int num_accounts = d->m_accounts.size();
-    int check_result = d->checkNVPairs(nv_pairs);
-    if (check_result < 0) { return check_result; }
 
     QList<uint> matching_accounts = findAccounts(nv_pairs);
     if (!matching_accounts.isEmpty()) { return matching_accounts[0]; }
 
-    uint account_handle = d->createAccount(d->ensurePresenceIn(nv_pairs));
+    QVariantMap data = d->sanityCheck(nv_pairs);
+    if (data.isEmpty()) { return 0; }
+    uint account_handle = d->createAccount(data);
+
+    if (account_handle != 0) { emit accountCreated(account_handle); }
 
     Q_ASSERT(d->m_accounts.size() == num_accounts + 1);
 
@@ -471,13 +522,38 @@ uint AccountManager::updateAccount(const uint account_handle,
 {
     int num_accounts = d->m_accounts.size();
 
-    int check_result = d->checkNVPairs(nv_pairs);
-    if (check_result < 0) { return 0; }
+    QVariantMap data = nv_pairs;
+    // Remove internal stuff:
+    data.remove(Decibel::name_current_presence);
+    int presence_state = data[Decibel::name_presence].toInt();
+    QString presence_message = data[Decibel::name_presence_message].toString();
+    data.remove(Decibel::name_presence);
+    data.remove(Decibel::name_presence_message);
 
-    d->m_accounts[account_handle] = d->ensurePresenceIn(nv_pairs);
+    // merge old and new data:
+    QVariantMap old_data = queryAccount(account_handle);
+    QString key;
+    foreach (key, old_data.keys())
+    {
+        if (!data.contains(key)) { data.insert(key, old_data[key]); }
+    }
+
+    // make sure the data is sane:
+    data = d->sanityCheck(data);
+
+    if (data.isEmpty()) { return 0; }
+
+    d->m_accounts[account_handle] = data;
+    if (presence_state != presence(account_handle) &&
+        presence_message != presenceMessage(account_handle))
+    { setPresenceAndMessage(account_handle, presence_state, presence_message); }
+    else if (presence_state != presence(account_handle))
+    { setPresence(account_handle, presence_state); }
+    else if (presence_message != presenceMessage(account_handle))
+    { setPresenceMessage(account_handle, presence_message); }
+    else
+    { emit accountUpdated(account_handle); }
 
-    Q_ASSERT(d->m_accounts[account_handle].data.contains(Decibel::name_protocol));
-    Q_ASSERT(d->m_accounts[account_handle].data.contains(Decibel::name_presence));
     Q_ASSERT(d->m_accounts.size() == num_accounts);
 
     return account_handle;
@@ -488,9 +564,11 @@ void AccountManager::deleteAccount(const uint account_handle)
     int num_accounts = d->m_accounts.size();
     int num_removals = d->m_accounts.remove(account_handle);
 
+    Q_ASSERT(num_removals == 0 || num_removals == 1);
     Q_ASSERT(num_removals != 0 || d->m_accounts.size() == num_accounts);
     Q_ASSERT(num_removals != 1 || d->m_accounts.size() == num_accounts - 1);
-    Q_ASSERT(num_removals >= 0 && num_removals <= 1);
+
+    if (num_removals == 1) { emit accountDeleted(account_handle); }
 }
 
 QList<uint> AccountManager::findAccounts(const QVariantMap & nv_pairs) const
@@ -528,6 +606,10 @@ int AccountManager::setPresence(const uint account_handle,
     { return Decibel::ERROR_NO_SUCH_ACCOUNT; }
 
     if (presence(account_handle) == presence_state) { return presence_state; }
+
+    d->m_accounts[account_handle].data[Decibel::name_presence] =
+        QVariant(presence_state);
+
     QString presence_message = presenceMessage(account_handle);
     int result = d->manageConnection(account_handle,
                                      QtTapioca::ContactBase::Presence(presence_state),
@@ -547,14 +629,36 @@ int AccountManager::presence(const uint account_handle)
     return result;
 }
 
+int AccountManager::currentPresence(const uint account_handle)
+{
+    if (!d->gotAccount(account_handle))
+    { return Decibel::ERROR_NO_SUCH_ACCOUNT; }
+
+    Q_ASSERT(d->m_accounts[account_handle].data.contains(Decibel::name_current_presence));
+    int result = d->m_accounts[account_handle].data[Decibel::name_current_presence].toInt();
+    Q_ASSERT(result >= QtTapioca::ContactBase::Offline &&
+             result <= QtTapioca::ContactBase::Busy);
+    return result;
+}
+
 int AccountManager::setPresenceMessage(const uint account_handle,
                                        const QString & message)
 {
     if (!d->gotAccount(account_handle))
     { return Decibel::ERROR_NO_SUCH_ACCOUNT; }
-    int result = d->manageConnection(account_handle,
+
+    int result = 0;
+    if (0 != d->connectionOf(account_handle))
+    {
+        result = d->manageConnection(account_handle,
                                      QtTapioca::ContactBase::Presence(presence(account_handle)),
                                      message);
+    }
+    else
+    {
+        // not connected: Just update our data:
+        d->m_accounts[account_handle].data[Decibel::name_presence_message] = message;
+    }
     return result;
 }
 
@@ -575,9 +679,17 @@ int AccountManager::setPresenceAndMessage(const uint account_handle,
 {
     if (!d->gotAccount(account_handle))
     { return Decibel::ERROR_NO_SUCH_ACCOUNT; }
-    return d->manageConnection(account_handle,
+    int result = 0;
+    result = d->manageConnection(account_handle,
                                QtTapioca::ContactBase::Presence(presence_state),
                                message);
+    if (presence_state == QtTapioca::ContactBase::Offline)
+    {
+        // not connected anymore, so we won't get any signals from telepathy
+        // about the message change...
+        d->m_accounts[account_handle].data[Decibel::name_presence_message] = message;
+    }
+    return result;
 }
 
 
@@ -600,11 +712,13 @@ void AccountManager::onOwnPresenceUpdated(QtTapioca::Connection * connection,
     int account_handle = d->findAccountWithConnection(connection);
     if (0 == account_handle) { return; }
 
+    Q_ASSERT(account_handle != 0);
     Q_ASSERT(d->gotAccount(account_handle));
 
-    d->updatePresence(account_handle,
-                      QtTapioca::ContactBase::Presence(presence_state),
-                      message);
+    d->updatePresenceLocally(account_handle,
+                             QtTapioca::ContactBase::Presence(presence_state),
+                             message);
+    emit accountUpdated(account_handle);
 }
 
 void AccountManager::onConnectionClosed(QtTapioca::Connection * connection)
@@ -614,8 +728,13 @@ void AccountManager::onConnectionClosed(QtTapioca::Connection * connection)
 
     Q_ASSERT(d->gotAccount(account_handle));
 
-    d->updatePresence(account_handle, QtTapioca::ContactBase::Offline, presenceMessage(account_handle));
+    d->updatePresenceLocally(account_handle, QtTapioca::ContactBase::Offline, presenceMessage(account_handle));
     d->setConnection(account_handle, 0);
 
-    Q_ASSERT(presence(account_handle) == QtTapioca::ContactBase::Offline);
+    emit accountUpdated(account_handle);
+
+    Q_ASSERT(currentPresence(account_handle) == QtTapioca::ContactBase::Offline);
 }
+
+void AccountManager::doBlockReconnects()
+{ d->blockReconnects = true; }
