@@ -58,6 +58,8 @@ namespace
     static const QString profiles_component_handle_key("Component Handle");
 }
 
+/// @cond INCLUDE_PRIVATE
+
 /**
  * @brief Private class implementing the ComponentManager.
  *
@@ -82,8 +84,6 @@ public:
         if (searchPath.isEmpty())
         { searchPath = component_search_path.split(";", QString::SkipEmptyParts); }
 
-        qDebug() << "ComponentManager: Search path set to" << searchPath;
-
         // Read Profiles:
         int num_profiles = settings.beginReadArray(profiles_array);
         for (int i = 0; i < num_profiles; ++i)
@@ -126,7 +126,7 @@ public:
                         {
                             int type = current.toInt();
                             if (type >= QtTapioca::Channel::Text &&
-                                type <= QtTapioca::Channel::Video)
+                                type <= QtTapioca::Channel::Stream)
                             { component.type_list << type; }
                         }
                     }
@@ -244,7 +244,6 @@ public:
         QString profile_name;
         foreach (profile_name, profiles.keys())
         {
-            qDebug() << "Fixing profile" << profile_name;
             int i = 0;
             while (i < profiles[profile_name].size())
             {
@@ -290,8 +289,6 @@ public:
     /** @brief Scan the search path for .component files and read them. */
     void scanComponents()
     {
-        qDebug() << "ComponentManager: Scanning for components";
-
         QHash<uint, Decibel::ComponentInfo> local_components;
 
         QString current_dir;
@@ -308,7 +305,7 @@ public:
                 continue;
             }
 
-            dir.setFilter(QDir::Files | QDir::NoSymLinks);
+            dir.setFilter(QDir::Files);
             dir.setNameFilters(QStringList("*.component"));
             dir.setSorting(QDir::Name);
 
@@ -337,7 +334,7 @@ public:
                         {
                             int type = current.toInt();
                             if (type >= QtTapioca::Channel::Text &&
-                                type <= QtTapioca::Channel::Video)
+                                type <= QtTapioca::Channel::Stream)
                             { info.possible_type_list << type; }
                         }
                     }
@@ -360,11 +357,20 @@ public:
                     else
                     {
                         qWarning() << "Unknown key" << key
-                                << "found while reading components.";
+                                << "found while reading component from"
+                                << fileprofile_name;
                     }
                 }
                 component_settings.endGroup();
 
+                // Do we have all the data we need?
+                if (info.service_name.isEmpty() ||
+                    info.object_path.path().isEmpty())
+                {
+                    qWarning() << fileprofile_name
+                               << "is incomplete, skipping.";
+                    continue;
+                }
                 // Is this component info already known?
                 bool already_known(false);
                 uint tmp_handle;
@@ -453,6 +459,8 @@ public:
     QStringList searchPath;
 };
 
+/// @endcond
+
 // ****************************************************************************
 
 ComponentManager::ComponentManager(QObject * parent) :
