@@ -29,12 +29,15 @@
 #include <QtCore/QCoreApplication>
 #include <QtCore/QPointer>
 #include <QtCore/QDebug>
+#include <QtCore/QTimer>
 
 #include <QtTapioca/Connection>
 #include <QtTapioca/Contact>
 #include <QtTapioca/ContactList>
 #include <QtTapioca/Handle>
 
+/// @cond INCLUDE_PRIVATE
+
 /**
  * @brief Private data class for the ContactManager.
  * @author Tobias Hunger <info@basyskom.de>
@@ -43,8 +46,10 @@ class ContactManagerPrivate : public QObject
 {
 public:
     /** @brief Constructor. */
-    ContactManagerPrivate(AccountManager * account_mgr) :
+    ContactManagerPrivate(AccountManager * account_mgr,
+                          ContactManager * contact_mgr) :
         accountManager(account_mgr),
+        contactManager(contact_mgr),
         m_connector(new SimplisticContactConnector())
     {
         Q_ASSERT(0 != accountManager);
@@ -54,6 +59,9 @@ public:
     /** @brief Destructor. */
     ~ContactManagerPrivate()
     {
+        const QtTapioca::Connection * connection;
+        foreach (connection, m_connectionlist.keys())
+        { deregisterContacts(connection); }
         delete m_connector;
     }
 
@@ -73,14 +81,29 @@ public:
         Q_ASSERT(0 != connection);
         Q_ASSERT(0 != contact);
         Q_ASSERT(!protocol.isEmpty());
-        QtTapioca::Handle * handle = contact->handle();
-        Q_ASSERT(0 != handle);
+
+        QString uri(contact->uri());
+        QString uri_protocol(getProtocol(uri));
+        if (uri_protocol.isEmpty())
+        { uri = protocol + QString("://") + uri; }
+        else
+        {
+            if (protocol != uri_protocol)
+            {
+                qWarning() << "CtM: Fixing protocol in URI, was" << protocol
+                           << "changed to" << uri_protocol;
+                uri = uri_protocol + "://" + getCMInternalUri(uri);
+            }
+        }
+        Q_ASSERT(uri.indexOf(QString("://") != 0));
+        Q_ASSERT(getProtocol(uri) == protocol);
+        Q_ASSERT(getCMInternalUri(uri) == getCMInternalUri(contact->uri()));
 
         // mark the handle as belonging to the given connection:
         m_connectionlist[connection].append(contact);
 
         // Update mapping between handles:
-        uint external_id = m_connector->findURI(protocol, contact->uri());
+        uint external_id = m_connector->findURI(uri);
         if (external_id == 0)
         {
             // FIXME: Add data to PIM?
@@ -110,6 +133,7 @@ public:
         QtTapioca::Contact * internal_contact = 0;
         foreach (internal_contact, m_connectionlist[connection])
         {
+            if (!m_telepathy2external.contains(internal_contact)) { continue; }
             Q_ASSERT(m_telepathy2external.contains(internal_contact));
             uint external_id = m_telepathy2external[internal_contact];
             Q_ASSERT(m_external2telepathy.contains(external_id));
@@ -117,8 +141,8 @@ public:
 
             m_external2telepathy.remove(external_id);
             m_telepathy2external.remove(internal_contact);
+            Q_ASSERT(m_telepathy2external.size() == m_external2telepathy.size());
         }
-        Q_ASSERT(m_telepathy2external.size() == m_external2telepathy.size());
         m_connectionlist.remove(connection);
     }
 
@@ -189,29 +213,161 @@ public:
         }
     }
 
-    /**
-     * @brief Check whether a contact can be reached using a given account.
-     * @param external_id The contact's ID in the external PIM system.
-     * @param account_handle A handle to an account.
-     * @return true if the contact can be reached from the account and false
-     * otherwise
-     *
-     * Check whether a contact can be reached using a given account. This
-     * does not include a check whether the contact is online. It is about
-     * whether there is any chance to ever see the contact from the account.
-     */
-    bool isReachable(const uint external_id, const int account_handle)
-    {
-        // FIXME: Implement this.
-        return true;
-    }
-
     /** @brief A pointer to the AccountManager. */
     QPointer<AccountManager> accountManager;
     /** @brief A pointer to the D-Bus Adaptor of the ContactManager. */
     QPointer<ContactManagerAdaptor> adaptor;
 
+    QStringList getURIs(const uint contact_id,
+                        const QString & protocol) const
+    { return m_connector->getURIs(contact_id, protocol); }
+
+    QtTapioca::Connection * connectAccount(const int account_handle)
+    {
+        // check Account:
+        if (!accountManager->gotAccount(account_handle))
+        {
+            qWarning() << "CtM: Account" << account_handle << "is invalid.";
+            return 0;
+        }
+
+        // Bring up account (if not already up):
+        if (accountManager->presence(account_handle) ==
+            QtTapioca::ContactBase::Offline)
+        {
+            Q_ASSERT(0 == accountManager->connectionOf(account_handle));
+            accountManager->setPresence(account_handle,
+                                        QtTapioca::ContactBase::Available);
+        }
+
+        // Wait for connection to get established...
+        QtTapioca::Connection * connection(0);
+        while (0 == connection)
+        {
+            QCoreApplication::instance()->processEvents();
+            connection = accountManager->connectionOf(account_handle);
+        }
+
+        return connection;
+    }
+
+    /** @brief Do Tapioca-Magic to set up the actual channel. */
+    Decibel::ChannelInfo
+    contactContactViaConnection(QtTapioca::Contact * const contact,
+                                QtTapioca::Connection * const connection,
+                                const int type,
+                                const bool suppress_handler)
+    {
+        // Make sure data is valid:
+        // * contact:
+        if (0 == contact)
+        {
+            qWarning() << "CtM: Contact not set up.";
+            return Decibel::ChannelInfo();
+        }
+
+        // * type:
+        if (type < int(QtTapioca::Channel::Text) ||
+            type > int(QtTapioca::Channel::Stream))
+        {
+            qWarning() << "CtM: Invalid type of connection requested:" << type;
+            return Decibel::ChannelInfo();
+        }
+
+        // * connection:
+        if (0 == connection)
+        {
+            qWarning() << "CtM: No connection set up.";
+            return Decibel::ChannelInfo();
+        }
+
+        // Do the work:
+        QtTapioca::Channel * channel =
+            connection->createChannel(QtTapioca::Channel::Type(type), contact,
+                                      suppress_handler);
+        if (0 == channel)
+        {
+            qWarning() << "CtM: Failed to create channel of type" << type
+                    << "to contact" << contact << "(suppress=" << suppress_handler
+                    << ").";
+            return Decibel::ChannelInfo();
+        }
+        qDebug() << "CtM: Got my channel!";
+
+        // decouple execution of channel handlers:
+        newChannelList.append(qMakePair(connection, channel));
+        QTimer::singleShot(10, contactManager, SLOT(onNewChannelCreated()));
+
+        Decibel::ChannelInfo channel_info(connection, channel);
+
+        Q_ASSERT(!channel_info.isNull());
+        return channel_info;
+    }
+
+    QString getProtocol(const QString & uri)
+    {
+        int pos = uri.indexOf(QString("://"));
+        if (pos < 0) { return QString(); }
+        else { return uri.left(pos); }
+    }
+
+    QString getCMInternalUri(const QString & uri)
+    {
+        int pos = uri.indexOf(QString("://"));
+        if (pos < 0) { return uri; }
+        else { return uri.mid(pos + 3); }
+    }
+
+    bool testUri(const QString & uri)
+    {
+        // Make sure contact_uri is valid:
+        QString remote_protocol = getProtocol(uri);
+        if (uri.isEmpty())
+        {
+            qWarning() << "CtM: Contact URI is empty.";
+            return false;
+        }
+        else if (remote_protocol.isEmpty())
+        {
+            qWarning() << "CtM: Contact URI" << uri
+                    << "does not name a protocol.";
+            return false;
+        }
+        return true;
+    }
+
+    /**
+     * @brief Find all accounts supporting a given protocol.
+     * @param protocol The protocol the account must support.
+     * @return A list of accounts supporting the requested protocol.
+     *
+     * Find all the accounts that support a given protocol. Accounts in an
+     * online state are returned first in the list.
+     */
+    QList<uint> findAccountsSupporting(const QString & protocol)
+    {
+        QList<uint> allAccounts(accountManager->listAccounts());
+        QList<uint> possibleAccounts;
+        QList<uint> possibleOnlineAccounts;
+        uint currentAccount;
+        foreach(currentAccount, allAccounts)
+        {
+            if (accountManager->protocol(currentAccount) == protocol)
+            {
+                if (0 != accountManager->connectionOf(currentAccount))
+                { possibleOnlineAccounts.append(currentAccount); }
+                else { possibleAccounts.append(currentAccount); }
+            }
+        }
+
+        return possibleOnlineAccounts + possibleAccounts;
+    }
+
+    QList<QPair<QtTapioca::Connection *, QtTapioca::Channel *> > newChannelList;
+
 private:
+    ContactManager * const contactManager;
+
     /** @brief A pointer to the ContactConnector used. */
     SimplisticContactConnector * m_connector;
 
@@ -230,14 +386,17 @@ private:
      */
     QHash<const QtTapioca::Connection *,
           QList<QtTapioca::Contact *> > m_connectionlist;
+
 };
 
+/// @endcond
+
 // **************************************************************************
 
 ContactManager::ContactManager(AccountManager * account_mgr,
                                QObject * parent) :
         QObject(parent),
-        d(new ContactManagerPrivate(account_mgr))
+        d(new ContactManagerPrivate(account_mgr, this))
 {
     Q_ASSERT(0 != d);
     d->adaptor = new ContactManagerAdaptor(this);
@@ -245,8 +404,48 @@ ContactManager::ContactManager(AccountManager * account_mgr,
 }
 
 ContactManager::~ContactManager()
+{ delete d; }
+
+Decibel::ChannelInfo
+ContactManager::contactUrlUsingAccount(const QString & contact_uri,
+                                       const int account_handle,
+                                       const int type,
+                                       const bool suppress_handler)
 {
-    delete d;
+    // Make sure contact_uri is valid:
+    if (!d->testUri(contact_uri))
+    { return Decibel::ChannelInfo(); }
+    QString remote_protocol = d->getProtocol(contact_uri);
+
+    QString local_protocol = d->accountManager->protocol(account_handle);
+    if (local_protocol != remote_protocol)
+    {
+        qWarning() << "CtM: Local protocol" << local_protocol
+                   << "and remote protocol" << remote_protocol
+                   << "differ!";
+    }
+
+    // This tests the account_handle:
+    QtTapioca::Connection * connection = d->connectAccount(account_handle);
+    if (0 == connection)
+    {
+        qWarning() << "CtM: Failed to setup Connection.";
+        return Decibel::ChannelInfo();
+    }
+
+    // We have a connection from this point on:
+    Q_ASSERT(0 != connection);
+
+    // Create contact:
+    QString cm_internal_uri = d->getCMInternalUri(contact_uri);
+    QtTapioca::Contact * contact(connection->contactList()->
+                                 contact(cm_internal_uri));
+    if (0 == contact)
+    { contact = connection->contactList()->addContact(cm_internal_uri); }
+    d->registerContact(remote_protocol, contact, connection);
+
+    return d->contactContactViaConnection(contact, connection,
+                                          type, suppress_handler);
 }
 
 Decibel::ChannelInfo
@@ -255,60 +454,21 @@ ContactManager::contactContactUsingAccount(const uint contact_id,
                                            const int type,
                                            const bool suppress_handler)
 {
-    // Make sure data is valid:
-    // * handles:
+    // Make sure contact handle is valid:
     if (!d->gotContact(contact_id))
     {
-        qWarning() << tr("Contact %1 is invalid.",
-                         "1: contact_id").arg(contact_id);
+        qWarning() << "CtM: Contact" << contact_id << "is invalid.";
         return Decibel::ChannelInfo();
     }
 
-    if (!d->accountManager->gotAccount(account_handle))
+    // This tests the account_handle:
+    QtTapioca::Connection * connection = d->connectAccount(account_handle);
+    if (0 == connection)
     {
-        qWarning() << tr("Account %1 is invalid.",
-                         "1: account_handle").arg(account_handle);
+        qWarning() << "CtM: Failed to setup Connection.";
         return Decibel::ChannelInfo();
     }
 
-    // * type:
-    if (type < int(QtTapioca::Channel::Text) ||
-        type > int(QtTapioca::Channel::Video))
-    {
-        qWarning() << tr("Invalid type of connection requested: %1.",
-                         "1: type").arg(type);
-        return Decibel::ChannelInfo();
-    }
-
-    // check whether contact is reachable from the account:
-    if (!d->isReachable(contact_id, account_handle))
-    {
-        qWarning() << tr("contact %1 can not be reached from account %2.",
-                         "1: contact_id, 2: account_handle").
-                      arg(contact_id).
-                      arg(account_handle);
-        return Decibel::ChannelInfo();
-    }
-
-    // Bring up account (if not already up):
-    if (d->accountManager->presence(account_handle) ==
-         QtTapioca::ContactBase::Offline)
-    {
-        Q_ASSERT(0 == d->accountManager->connectionOf(account_handle));
-        d->accountManager->setPresence(account_handle,
-                                       QtTapioca::ContactBase::Available);
-    }
-
-    // Wait for connection to get established...
-    QtTapioca::Connection * connection(0);
-    qDebug(qPrintable(tr("Looping for connection...")));
-    while (0 == connection)
-    {
-        QCoreApplication::instance()->processEvents();
-        connection = d->accountManager->connectionOf(account_handle);
-    }
-    qDebug(qPrintable(tr("Connection found.")));
-
     // We have a connection from this point on:
     Q_ASSERT(0 != connection);
 
@@ -316,20 +476,46 @@ ContactManager::contactContactUsingAccount(const uint contact_id,
     QtTapioca::Contact * contact = d->mapFromExternal(contact_id);
     if (0 == contact)
     {
-        qWarning() << tr("Can not open a channel to contact %1: Not online.",
-                         "1: contact_id").arg(contact_id);
-        return Decibel::ChannelInfo();
+        QString protocol(d->accountManager->protocol(account_handle));
+        QStringList uris = d->getURIs(contact_id, protocol);
+        if (uris.isEmpty())
+        {
+            qWarning() << "CtM: Contact" << contact_id
+                       << "has no support for protocol" << protocol;
+            return Decibel::ChannelInfo();
+        }
+        // FIXME: Handle multiple uris...
+        contact = connection->contactList()->addContact(d->getCMInternalUri(uris[0]));
+        d->registerContact(protocol, contact, connection);
+        Q_ASSERT(d->mapFromExternal(contact_id) == contact);
     }
-    qDebug(qPrintable(tr("Contact is online.")));
 
-    QtTapioca::Channel * channel =
-        connection->createChannel(QtTapioca::Channel::Type(type), contact,
-                                  suppress_handler);
-    Q_ASSERT(channel != 0);
-
-    return Decibel::ChannelInfo(connection, channel);
+    return d->contactContactViaConnection(contact, connection,
+                                          type, suppress_handler);
 }
 
+Decibel::ChannelInfo
+ContactManager::contactUrl(const QString & contact_uri, const int type,
+                           const bool suppress_handler)
+{
+    // Make sure contact_uri is valid:
+    if (!d->testUri(contact_uri))
+    { return Decibel::ChannelInfo(); }
+    QString remote_protocol = d->getProtocol(contact_uri);
+
+    // Find accounts supporting protocol:
+    QList<uint> possibleAccounts = d->findAccountsSupporting(remote_protocol);
+
+    if (possibleAccounts.isEmpty())
+    {
+        qWarning() << "CtM: No account supporting" << remote_protocol
+                   << "found.";
+        return Decibel::ChannelInfo();
+    }
+
+    return contactUrlUsingAccount(contact_uri, possibleAccounts[0], type,
+                                  suppress_handler);
+}
 
 void ContactManager::onConnectionOpened(QtTapioca::Connection * connection)
 {
@@ -378,7 +564,8 @@ void ContactManager::onConnectionOpened(QtTapioca::Connection * connection)
 void ContactManager::onConnectionClosed(QtTapioca::Connection * connection)
 {
     Q_ASSERT(0 != connection);
-    Q_ASSERT(connection->status() == QtTapioca::Connection::Disconnected);
+    // This can hit all connections due to a crash, etc. so it does not
+    // make sense to assert the status of the connection here.
 
     d->deregisterContacts(connection);
 }
@@ -408,3 +595,11 @@ void ContactManager::onPresenceUpdated(QtTapioca::ContactBase * internal_contact
             qWarning() << "Unknown Handle type encountered.";
     }
 }
+
+void ContactManager::onNewChannelCreated()
+{
+    QPair<QtTapioca::Connection *, QtTapioca::Channel *> pair;
+    foreach (pair, d->newChannelList)
+    { emit channelOpened(pair.first, pair.second); }
+    d->newChannelList.clear();
+}
